import pathlib
from subprocess import list2cmdline
from typing import Optional, Tuple

import click

import anyscale
from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client import HaJobStates
from anyscale.controllers.job_controller import JobController
from anyscale.job.models import JobConfig
from anyscale.util import validate_non_negative_arg


log = BlockLogger()  # CLI Logger


@click.group("job", help="Interact with production jobs running on Anyscale.")
def job_cli() -> None:
    pass


@job_cli.command(
    name="submit", short_help="Submit a job.",
)
@click.option("--name", "-n", required=False, default=None, help="Name of the job.")
@click.option(
    "--description", required=False, default=None, help="Description of the job."
)
@click.option(
    "--follow",
    "-f",
    required=False,
    default=False,
    type=bool,
    is_flag=True,
    help="Whether to follow the log of the created job",
)
@click.option(
    "--wait",
    "-w",
    required=False,
    default=False,
    type=bool,
    is_flag=True,
    help="Whether to block this CLI command until the job succeeds (or reaches another terminal state)",
)
@click.argument("entrypoint", required=False, nargs=-1, type=click.UNPROCESSED)
def submit(
    entrypoint: Tuple[str],
    name: Optional[str],
    description: Optional[str],
    follow: Optional[bool],
    wait: Optional[bool],
):
    """Submit a job.

    The job config can be specified from a file or inlined in this command:

    $ anyscale job submit config.yaml # Submit from a config file.

    $ anyscale job submit -- python main.py # Submit from an inlined entrypoint.

    By default, this command submits the job asynchronously and exits. To wait for the job to complete, use the `--wait` flag.
    """

    if len(entrypoint) == 0:
        raise click.ClickException(
            "Either a config file or an inlined entrypoint must be provided."
        )

    config_path: Optional[str] = None
    if len(entrypoint) == 1:
        maybe_config_path = entrypoint[0]
        if pathlib.Path(maybe_config_path).is_file():
            config_path = entrypoint[0]
            log.info(f"Submitting job from config file: '{config_path}'.")
        elif maybe_config_path.endswith(".yaml"):
            raise click.ClickException(
                f"Job config file '{maybe_config_path}' not found."
            )

    job_controller = JobController()
    if config_path is not None:
        job_id = job_controller.submit(config_path, name=name, description=description,)
    else:
        job_id = anyscale.job.submit(
            JobConfig(name=name, entrypoint=list2cmdline(entrypoint))
        )

    if follow or wait:
        job_controller.log.info(
            "Waiting for the job to run. Interrupting this command will not cancel the job."
        )
    else:
        job_controller.log.info(
            "Use `--wait` to wait for the job to run and stream logs."
        )

    if follow:
        job_controller.logs(job_id, should_follow=follow)
    if wait:
        job_controller.wait(job_id=job_id)


@job_cli.command(name="list", help="Display information about existing jobs.")
@click.option("--name", "-n", required=False, default=None, help="Filter by job name.")
@click.option(
    "--job-id", "--id", required=False, default=None, help="Filter by job id."
)
@click.option(
    "--project-id", required=False, default=None, help="Filter by project id."
)
@click.option(
    "--include-all-users",
    is_flag=True,
    default=False,
    help="Include jobs not created by current user.",
)
@click.option(
    "--include-archived",
    is_flag=True,
    default=False,
    help=(
        "List archived jobs as well as unarchived jobs."
        "If not provided, defaults to listing only unarchived jobs."
    ),
)
@click.option(
    "--max-items",
    required=False,
    default=10,
    type=int,
    help="Max items to show in list.",
    callback=validate_non_negative_arg,
)
def list(  # noqa: A001
    name: Optional[str],
    job_id: Optional[str],
    project_id: Optional[str],
    include_all_users: bool,
    include_archived: bool,
    max_items: int,
) -> None:
    job_controller = JobController()
    job_controller.list(
        name=name,
        job_id=job_id,
        project_id=project_id,
        include_all_users=include_all_users,
        include_archived=include_archived,
        max_items=max_items,
    )


@job_cli.command(name="archive", help="Archive a job.")
@click.option("--job-id", "--id", required=False, help="Unique ID of the job.")
@click.option("--name", "-n", required=False, help="Name of the job.")
def archive(job_id: Optional[str], name: Optional[str]) -> None:
    job_controller = JobController()
    job_controller.archive(job_id=job_id, job_name=name)


@job_cli.command(name="terminate", help="Attempt to terminate a job asynchronously.")
@click.option("--job-id", "--id", required=False, help="Unique ID of the job.")
@click.option("--name", "-n", required=False, help="Name of the job.")
def terminate(job_id: Optional[str], name: Optional[str]) -> None:
    job_controller = JobController()
    job_controller.terminate(job_id=job_id, job_name=name)


@job_cli.command(name="logs")
@click.option("--job-id", "--id", required=False, help="Unique ID of the job.")
@click.option("--name", "-n", required=False, help="Name of the job.")
@click.option(
    "--follow",
    "-f",
    required=False,
    default=False,
    type=bool,
    is_flag=True,
    help="Whether to follow the log.",
)
@click.option(
    "--all-attempts",
    is_flag=True,
    default=False,
    help="Show logs for all job attempts.",
)
def logs(
    job_id: Optional[str],
    name: Optional[str],
    follow: bool = False,
    all_attempts: bool = False,
) -> None:
    """Print the logs of a job.

    By default from the latest job attempt.

    Example usage:

        anyscale job logs --id prodjob_123

        anyscale job logs --id prodjob_123 -f

        anyscale job logs -n my-job --all-attempts"""
    job_controller = JobController(raise_structured_exception=True)
    job_controller.logs(
        job_id=job_id, job_name=name, should_follow=follow, all_attempts=all_attempts,
    )


@job_cli.command(name="wait")
@click.option("--job-id", "--id", required=False, help="Unique ID of the job.")
@click.option("--name", "-n", required=False, help="Name of the job.")
@click.option(
    "--state",
    "-s",
    required=False,
    default=HaJobStates.SUCCESS,
    help="The state to wait for this Job to enter",
)
@click.option(
    "--timeout",
    "-t",
    required=False,
    default=None,
    type=float,
    help="The timeout in seconds after which this command will exit.",
)
def wait(
    job_id: Optional[str],
    name: Optional[str],
    state: str = HaJobStates.SUCCESS,
    timeout=None,
) -> None:
    """Wait for a Job to enter a specific state (default: SUCCESS).

    To specify the Job by name, use the --name flag. To specify the Job by id, use the --job-id flag.

    If the Job reaches the target state, the command will exit successfully.

    If the Job reaches a terminal state other than the target state, the command will exit with an error.

    If the command reaches the timeout, the command will exit with an error.
    """
    state = state.upper()
    job_controller = JobController()
    job_id = job_controller.wait(
        job_id=job_id, job_name=name, target_state=state, timeout_secs=timeout,
    )



import copy
import os
from typing import Any, Dict, Generator, Optional

import click
from click.testing import CliRunner
import pytest

from anyscale._private.sdk import _LAZY_SDK_SINGLETONS
from anyscale.commands.service_commands import deploy, status
from anyscale.service.commands import _SERVICE_SDK_SINGLETON_KEY
from anyscale.service.models import (
    ServiceConfig,
    ServiceState,
    ServiceStatus,
    ServiceVersionStatus,
)


def _get_test_file_path(subpath: str) -> str:
    return os.path.join(os.path.dirname(__file__), "test_files/", subpath,)


EMPTY_CONFIG_PATH = _get_test_file_path("service_config_files/empty.yaml")
MINIMAL_CONFIG_PATH = _get_test_file_path("service_config_files/minimal.yaml")
FULL_CONFIG_PATH = _get_test_file_path("service_config_files/full.yaml")
NAME_AND_GIBBERISH_CONFIG_PATH = _get_test_file_path(
    "service_config_files/name_and_gibberish.yaml"
)
MULTI_LINE_REQUIREMENTS_PATH = _get_test_file_path("requirements_files/multi_line.txt")


class FakeServiceSDK:
    DEFAULT_SERVICE_ID = "default-fake-service-id"
    DEFAULT_SERVICE_NAME = "default-fake-service-name"

    def __init__(self):
        self._deployed_config: Optional[ServiceConfig] = None
        self._deployed_kwargs: Dict[str, Any] = {}
        self._fetched_name: Optional[str] = None

    @property
    def deployed_config(self) -> Optional[ServiceConfig]:
        return self._deployed_config

    @property
    def deployed_kwargs(self) -> Dict[str, Any]:
        return copy.deepcopy(self._deployed_kwargs)

    def deploy(self, config: ServiceConfig, **kwargs):
        assert isinstance(config, ServiceConfig)
        self._deployed_config = config
        self._deployed_kwargs = kwargs

    @property
    def fetched_name(self) -> Optional[str]:
        return self._fetched_name

    def status(self, name: Optional[str] = None) -> ServiceStatus:
        self._fetched_name = name
        return ServiceStatus(
            id=self.DEFAULT_SERVICE_ID,
            name=name or self.DEFAULT_SERVICE_NAME,
            state=ServiceState.TERMINATED,
            query_url="http://fake-service-url/",
            query_auth_token="asdf1234",
            primary_version=ServiceVersionStatus(
                name="primary",
                weight=100,
                config=ServiceConfig(applications=[{"import_path": "main:app"}]),
            ),
            canary_version=ServiceVersionStatus(
                name="canary",
                weight=0,
                config=ServiceConfig(applications=[{"import_path": "main:app"}]),
            ),
        )


@pytest.fixture()
def fake_service_sdk() -> Generator[FakeServiceSDK, None, None]:
    fake_service_sdk = FakeServiceSDK()
    _LAZY_SDK_SINGLETONS[_SERVICE_SDK_SINGLETON_KEY] = fake_service_sdk
    try:
        yield fake_service_sdk
    finally:
        del _LAZY_SDK_SINGLETONS[_SERVICE_SDK_SINGLETON_KEY]


def _assert_error_message(result: click.testing.Result, *, message: str):
    assert result.exit_code != 0
    assert message in result.stdout


class TestDeploy:
    def test_deploy_no_arg(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(deploy)
        _assert_error_message(
            result, message="Either config file or import path must be provided."
        )

    def test_deploy_from_import_path(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(deploy, ["main:app"])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(applications=[{"import_path": "main:app"}])
        assert (
            fake_service_sdk.deployed_config.applications
            == expected_config.applications
        )

    def test_deploy_from_import_path_with_args(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(deploy, ["main:app", "arg1=val1", "arg2=val2"])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            applications=[
                {
                    "import_path": "main:app",
                    "arguments": {"arg1": "val1", "arg2": "val2"},
                }
            ],
        )
        assert (
            fake_service_sdk.deployed_config.applications
            == expected_config.applications
        )

    def test_deploy_from_import_path_with_bad_arg(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(deploy, ["main:app", "bad_arg"])
        _assert_error_message(result, message="Invalid application argument 'bad_arg'")

    def test_deploy_from_file(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(deploy, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(applications=[{"import_path": "main:app"}])
        assert (
            fake_service_sdk.deployed_config.applications
            == expected_config.applications
        )

    def test_deploy_from_file_with_import_path(self, fake_service_sdk):
        runner = CliRunner()
        os.path.join(
            os.path.dirname(__file__), "test_service_config_files", "minimal.yaml",
        )

        result = runner.invoke(
            deploy, ["-f", MINIMAL_CONFIG_PATH, "main:app", "arg1=val1"]
        )
        _assert_error_message(
            result,
            message="When a config file is provided, import path and application arguments can't be.",
        )

        result = runner.invoke(deploy, ["-f", MINIMAL_CONFIG_PATH, "main:app"])
        _assert_error_message(
            result,
            message="When a config file is provided, import path and application arguments can't be.",
        )

    @pytest.mark.parametrize("flag", ["--in-place", "-i"])
    def test_deploy_in_place(self, fake_service_sdk, flag: str):
        runner = CliRunner()
        result = runner.invoke(deploy, ["main:app", flag])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(applications=[{"import_path": "main:app"}])
        assert (
            fake_service_sdk.deployed_config.applications
            == expected_config.applications
        )
        assert fake_service_sdk.deployed_kwargs == {
            "canary_percent": None,
            "in_place": True,
            "max_surge_percent": None,
        }

    def test_deploy_canary_percent(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(deploy, ["main:app", "--canary-percent", "50"])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(applications=[{"import_path": "main:app"}])
        assert (
            fake_service_sdk.deployed_config.applications
            == expected_config.applications
        )
        assert fake_service_sdk.deployed_kwargs == {
            "canary_percent": 50,
            "in_place": False,
            "max_surge_percent": None,
        }

    def test_deploy_max_surge_percent(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(deploy, ["main:app", "--max-surge-percent", "50"])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(applications=[{"import_path": "main:app"}])
        assert (
            fake_service_sdk.deployed_config.applications
            == expected_config.applications
        )
        assert fake_service_sdk.deployed_kwargs == {
            "canary_percent": None,
            "in_place": False,
            "max_surge_percent": 50,
        }

    def test_deploy_excludes(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(deploy, ["main:app"])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(applications=[{"import_path": "main:app"}])
        assert (
            fake_service_sdk.deployed_config.applications
            == expected_config.applications
        )

        # Pass a single exclusion.
        result = runner.invoke(deploy, ["--exclude", "path1", "main:app"])
        assert result.exit_code == 0, result.stdout
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            applications=[{"import_path": "main:app"}], excludes=["path1"]
        )
        assert (
            fake_service_sdk.deployed_config.applications
            == expected_config.applications
        )

        # Pass multiple exclusions.
        result = runner.invoke(
            deploy, ["--exclude", "path1", "-e", "path2", "main:app"]
        )
        assert result.exit_code == 0, result.stdout
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            applications=[{"import_path": "main:app"}], excludes=["path1", "path2"]
        )
        assert (
            fake_service_sdk.deployed_config.applications
            == expected_config.applications
        )

    def test_deploy_from_file_override_options(self, fake_service_sdk):
        runner = CliRunner()

        # No overrides, should match the config in the file.
        result = runner.invoke(deploy, ["-f", FULL_CONFIG_PATH])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-name-from-file",
            image_uri="docker.io/library/test:latest",
            compute_config="test-compute-config",
            working_dir="test-working-dir",
            excludes=["test"],
            requirements=["pip-install-test"],
            query_auth_token_enabled=False,
            http_options={"request_timeout_s": 10.0},
            grpc_options={"grpc_servicer_functions": ["hello.world"],},
            logging_config={"log_level": "DEBUG"},
        )
        assert fake_service_sdk.deployed_config == expected_config

        # Override name.
        result = runner.invoke(
            deploy, ["-f", FULL_CONFIG_PATH, "--name", "override-name"]
        )
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="override-name",
            image_uri="docker.io/library/test:latest",
            compute_config="test-compute-config",
            working_dir="test-working-dir",
            excludes=["test"],
            requirements=["pip-install-test"],
            query_auth_token_enabled=False,
            http_options={"request_timeout_s": 10.0},
            grpc_options={"grpc_servicer_functions": ["hello.world"],},
            logging_config={"log_level": "DEBUG"},
        )
        assert fake_service_sdk.deployed_config == expected_config

        # Override image URI.
        result = runner.invoke(
            deploy,
            [
                "-f",
                FULL_CONFIG_PATH,
                "--image-uri",
                "docker.io/user/override-image:latest",
            ],
        )
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-name-from-file",
            image_uri="docker.io/user/override-image:latest",
            compute_config="test-compute-config",
            working_dir="test-working-dir",
            excludes=["test"],
            requirements=["pip-install-test"],
            query_auth_token_enabled=False,
            http_options={"request_timeout_s": 10.0},
            grpc_options={"grpc_servicer_functions": ["hello.world"],},
            logging_config={"log_level": "DEBUG"},
        )
        assert fake_service_sdk.deployed_config == expected_config

        # Override compute_config.
        result = runner.invoke(
            deploy,
            ["-f", FULL_CONFIG_PATH, "--compute-config", "override-compute-config"],
        )
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-name-from-file",
            image_uri="docker.io/library/test:latest",
            compute_config="override-compute-config",
            working_dir="test-working-dir",
            excludes=["test"],
            requirements=["pip-install-test"],
            query_auth_token_enabled=False,
            http_options={"request_timeout_s": 10.0},
            grpc_options={"grpc_servicer_functions": ["hello.world"],},
            logging_config={"log_level": "DEBUG"},
        )
        assert fake_service_sdk.deployed_config == expected_config

        # Override working_dir.
        result = runner.invoke(
            deploy, ["-f", FULL_CONFIG_PATH, "--working-dir", "override-working-dir"]
        )
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-name-from-file",
            image_uri="docker.io/library/test:latest",
            compute_config="test-compute-config",
            working_dir="override-working-dir",
            excludes=["test"],
            requirements=["pip-install-test"],
            query_auth_token_enabled=False,
            http_options={"request_timeout_s": 10.0},
            grpc_options={"grpc_servicer_functions": ["hello.world"],},
            logging_config={"log_level": "DEBUG"},
        )
        assert fake_service_sdk.deployed_config == expected_config

        # Override requirements.
        result = runner.invoke(
            deploy,
            ["-f", FULL_CONFIG_PATH, "--requirements", MULTI_LINE_REQUIREMENTS_PATH],
        )
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-name-from-file",
            image_uri="docker.io/library/test:latest",
            compute_config="test-compute-config",
            working_dir="test-working-dir",
            excludes=["test"],
            requirements=MULTI_LINE_REQUIREMENTS_PATH,
            query_auth_token_enabled=False,
            http_options={"request_timeout_s": 10.0},
            grpc_options={"grpc_servicer_functions": ["hello.world"],},
            logging_config={"log_level": "DEBUG"},
        )
        assert fake_service_sdk.deployed_config == expected_config


class TestStatus:
    def test_no_name(self, fake_service_sdk):
        # Test passing no name (would be picked up as workspace name).
        runner = CliRunner()
        result = runner.invoke(status)
        assert result.exit_code == 0
        assert fake_service_sdk.fetched_name is None

    def test_name_and_config_file(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(status, ["-n", "custom-name", "-f", "file-name.yaml"])
        _assert_error_message(
            result, message="Only one of '--name' and '--config-file' can be provided."
        )

    def test_name(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(status, ["-n", "custom-name"])
        assert result.exit_code == 0
        assert fake_service_sdk.fetched_name == "custom-name"

    def test_service_config_file_not_found(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(status, ["-f", "nonexistent.yaml"])
        _assert_error_message(
            result, message="Config file not found at path: 'nonexistent.yaml'."
        )

    def test_empty_config_file(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(status, ["-f", EMPTY_CONFIG_PATH])
        _assert_error_message(
            result,
            message=f"No 'name' property found in config file '{EMPTY_CONFIG_PATH}'.",
        )

    def test_service_config_file_missing_name(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(status, ["-f", MINIMAL_CONFIG_PATH])
        _assert_error_message(
            result,
            message=f"No 'name' property found in config file '{MINIMAL_CONFIG_PATH}'.",
        )

    def test_service_config_file(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(status, ["-f", FULL_CONFIG_PATH])
        assert result.exit_code == 0
        assert fake_service_sdk.fetched_name == "test-name-from-file"

    def test_service_config_file_only_reads_name(self, fake_service_sdk):
        # `status` CLI path should ignore any other fields in the config file aside from `name`.
        runner = CliRunner()
        result = runner.invoke(status, ["-f", NAME_AND_GIBBERISH_CONFIG_PATH])
        assert result.exit_code == 0
        assert fake_service_sdk.fetched_name == "service-name-from-file"

    def test_verbose_flag(self, fake_service_sdk):
        runner = CliRunner()

        # No verbose flag -- exclude details.
        result = runner.invoke(status, ["-n", "custom-name"])
        assert result.exit_code == 0
        assert fake_service_sdk.fetched_name == "custom-name"
        assert "primary_version" in result.stdout
        assert "canary_version" in result.stdout
        assert "query_url" not in result.stdout
        assert "query_auth_token" not in result.stdout
        assert "config" not in result.stdout
        assert "applications" not in result.stdout

        # Verbose flag -- include details.
        result = runner.invoke(status, ["-n", "custom-name", "-v"])
        assert result.exit_code == 0
        assert fake_service_sdk.fetched_name == "custom-name"
        assert "primary_version" in result.stdout
        assert "canary_version" in result.stdout
        assert "query_url" in result.stdout
        assert "query_auth_token" in result.stdout
        assert "config" in result.stdout
        assert "applications" in result.stdout



from dataclasses import dataclass, field
from typing import Any, Dict, Optional

from anyscale._private.models import WorkloadConfig


@dataclass(frozen=True)
class JobConfig(WorkloadConfig):
    entrypoint: str = field(default="", repr=False)

    def _validate_entrypoint(self, entrypoint: str):
        if not isinstance(entrypoint, str):
            raise TypeError("'entrypoint' must be a string.")

        if not entrypoint:
            raise ValueError("'entrypoint' cannot be empty.")

    max_retries: int = field(default=1, repr=False)

    def _validate_max_retries(self, max_retries: int):
        if not isinstance(max_retries, int):
            raise TypeError("'max_retries' must be an int.")

        if max_retries < 0:
            raise ValueError("'max_retries' must be >= 0.")

    runtime_env: Optional[Dict[str, Any]] = field(default=None, repr=False)

    def _validate_runtime_env(self, runtime_env: Optional[Dict[str, Any]]):
        if runtime_env is None:
            return

        if not isinstance(runtime_env, dict):
            raise TypeError("'runtime_env' must be a dictionary.")



import asyncio
import re
from typing import List, Optional

import aiohttp

from anyscale.shared_anyscale_utils.utils.asyncio import gather_in_batches


CLUSTER_CONNECT_TIMEOUT = 30


async def _download_logs_concurrently(
    log_chunk_urls: List[str], parallelism: int, bearer_token: Optional[str] = None
) -> str:
    logs_across_chunks: List[str] = await gather_in_batches(  # type: ignore
        parallelism,
        *[
            _download_log_from_s3_url(url, bearer_token=bearer_token)
            for url in log_chunk_urls
        ],
    )
    logs_across_chunks = [log.strip() for log in logs_across_chunks]
    return "\n".join(logs_across_chunks)


async def _download_log_from_ray_json_response(url: str) -> str:
    async with aiohttp.ClientSession() as session:
        response = await asyncio.wait_for(
            session.get(url), timeout=CLUSTER_CONNECT_TIMEOUT
        )
        logs: str = (await response.json()).get("logs", "")
        return logs


async def _download_log_from_s3_url(
    url: str, bearer_token: Optional[str] = None,
) -> str:
    # Note that the URL is presigned, so no token needs to be passed in the request
    async with aiohttp.ClientSession() as session:
        headers = {"Authorization": f"Bearer {bearer_token}"} if bearer_token else {}
        async with session.get(url, headers=headers) as response:
            return await response.text()


def _remove_ansi_escape_sequences(s: str) -> str:
    # Required as the log may contain ANSI escape sequeneces (e.g. for coloring in the terminal)
    # Regex pattern from https://stackoverflow.com/a/14693789
    ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
    return ansi_escape.sub("", s)



from unittest.mock import Mock, patch

import click
import pytest

from anyscale.cluster import get_job_submission_client_cluster_info


@pytest.mark.parametrize(
    "mock_sessions_return_value",
    [
        Mock(results=[]),
        Mock(
            results=[Mock(host_name="mock_host_name", access_token="mock_access_token")]
        ),
    ],
)
@pytest.mark.parametrize("create_cluster_if_needed", [True, False])
def test_get_job_submission_client_cluster_info(
    mock_sessions_return_value: Mock, create_cluster_if_needed: bool
):
    mock_project_definition = Mock()
    mock_project_definition.root = "/some/directory"

    mock_api_client = Mock()
    mock_api_client.get_user_info_api_v2_userinfo_get = Mock(
        return_value=Mock(result=Mock(id="mock_user_id"))
    )
    mock_api_client.list_sessions_api_v2_sessions_get = Mock(
        return_value=mock_sessions_return_value
    )
    mock_project_block = Mock(return_value=Mock(project_id="mock_project_id"))
    mock_prepare_cluster_block = Mock(
        return_value=Mock(cluster_name="mock_cluster_name")
    )

    class MockClientBuilder:
        def __init__(self, **kwargs):
            self._project_dir = "_project_dir"
            self._project_name = "_project_name"
            self._cluster_name = "mock_cluster_name"
            self._autosuspend_timeout = ("_autosuspend_timeout",)
            self._allow_public_internet_traffic = "_allow_public_internet_traffic"
            self._needs_update = "_needs_update"
            self._cluster_compute_name = "_cluster_compute_name"
            self._cluster_compute_dict = "_cluster_compute_dict"
            self._cloud_name = "_cloud_name"
            self._build_pr = "_build_pr"
            self._force_rebuild = "_force_rebuild"
            self._build_commit = "_build_commit"
            self._cluster_env_name = "_cluster_env_name"
            self._cluster_env_dict = "_cluster_env_dict"
            self._cluster_env_revision = "_cluster_env_revision"
            self._ray = "_ray"
            self._cluster_compute_version = 2

    mock_client_builder = MockClientBuilder()

    with patch.multiple(
        "anyscale.cluster",
        ClientBuilder=MockClientBuilder,
        create_prepare_cluster_block=mock_prepare_cluster_block,
        create_project_block=mock_project_block,
        get_auth_api_client=Mock(return_value=Mock(api_client=mock_api_client)),
    ):
        if len(mock_sessions_return_value.results) == 0:
            with pytest.raises(click.ClickException):
                get_job_submission_client_cluster_info(
                    "mock_cluster_name", create_cluster_if_needed
                )
        else:
            get_job_submission_client_cluster_info(
                "mock_cluster_name", create_cluster_if_needed
            )

    mock_api_client.list_sessions_api_v2_sessions_get.assert_called_once_with(
        project_id="mock_project_id", active_only=True, name="mock_cluster_name"
    )

    if create_cluster_if_needed:
        mock_project_block.assert_called_once_with(
            mock_client_builder._project_dir,
            mock_client_builder._project_name,
            cloud_name=mock_client_builder._cloud_name,
            cluster_compute_name=mock_client_builder._cluster_compute_name,
            cluster_compute_dict=mock_client_builder._cluster_compute_dict,
        )
        mock_prepare_cluster_block.assert_called_once_with(
            project_id="mock_project_id",
            cluster_name=mock_client_builder._cluster_name,
            autosuspend_timeout=mock_client_builder._autosuspend_timeout,
            allow_public_internet_traffic=mock_client_builder._allow_public_internet_traffic,
            needs_update=mock_client_builder._needs_update,
            cluster_compute_name=mock_client_builder._cluster_compute_name,
            cluster_compute_dict=mock_client_builder._cluster_compute_dict,
            cloud_name=mock_client_builder._cloud_name,
            build_pr=mock_client_builder._build_pr,
            force_rebuild=mock_client_builder._force_rebuild,
            build_commit=mock_client_builder._build_commit,
            cluster_env_name=mock_client_builder._cluster_env_name,
            cluster_env_dict=mock_client_builder._cluster_env_dict,
            cluster_env_revision=mock_client_builder._cluster_env_revision,
            ray=mock_client_builder._ray,
            cluster_compute_version=mock_client_builder._cluster_compute_version,
        )



from dataclasses import dataclass
import os
from typing import List, Optional

from anyscale.cli_logger import BlockLogger
from anyscale.sdk.anyscale_client.configuration import Configuration


OPENAPI_NO_VALIDATION = Configuration()
OPENAPI_NO_VALIDATION.client_side_validation = False

TEST_COMPUTE_CONFIG_DICT = {
    "cloud_id": "fake-cloud-id",
    "region": "fake-region",
    "head_node_type": {
        "name": "fake-head-node-name",
        "instance_type": "fake-instance-type",
    },
    "worker_node_types": [],
    "auto_select_worker_config": False,
}

TEST_CONTAINERFILE = os.path.join(
    os.path.dirname(__file__), "test_files/containerfiles", "Containerfile",
)


@dataclass
class RequirementsFile:
    name: str
    expected_pip_list: Optional[List[str]]

    def get_path(self) -> str:
        return os.path.join(
            os.path.dirname(__file__), "test_files/requirements_files", self.name,
        )


SINGLE_LINE_REQUIREMENTS = RequirementsFile("single_line.txt", ["pip-install-test"])
MULTI_LINE_REQUIREMENTS = RequirementsFile(
    "multi_line.txt", ["pip-install-test", "torch==1.10.1"]
)

TEST_REQUIREMENTS_FILES = [
    SINGLE_LINE_REQUIREMENTS,
    MULTI_LINE_REQUIREMENTS,
    RequirementsFile("does_not_exist.txt", None),
    RequirementsFile("empty.txt", []),
    RequirementsFile(
        "multi_line_with_whitespace.txt",
        ["pip-install-test", "torch==1.10.1", "something-else"],
    ),
    RequirementsFile("comments.txt", ["pip-install-test", "torch==1.10.1"]),
]


class TestLogger:
    def __init__(self):
        self._delegate = BlockLogger()
        self._info_messages: List[str] = []
        self._warning_messages: List[str] = []
        self._error_messages: List[str] = []

    @property
    def info_messages(self) -> List[str]:
        return self._info_messages

    def info(self, msg: str):
        self._info_messages.append(msg)
        self._delegate.info(msg)

    @property
    def warning_messages(self) -> List[str]:
        return self._warning_messages

    def warning(self, msg: str):
        self._warning_messages.append(msg)
        self._delegate.warning(msg)

    @property
    def error_messages(self) -> List[str]:
        return self._error_messages

    def error(self, msg: str):
        self._error_messages.append(msg)
        self._delegate.error(msg)



import re
from typing import Optional, Tuple

from anyscale.authenticate import get_auth_api_client
from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.api.default_api import DefaultApi
from anyscale.cloud import get_cloud_id_and_name, get_last_used_cloud
from anyscale.sdk.anyscale_client import (
    ArchiveStatus,
    ComputeTemplateConfig,
    CreateComputeTemplate,
)
from anyscale.sdk.anyscale_client.models.cluster_compute_config import (
    ClusterComputeConfig,
)
from anyscale.sdk.anyscale_client.models.compute_template import ComputeTemplate
from anyscale.sdk.anyscale_client.models.compute_template_query import (
    ComputeTemplateQuery,
)
from anyscale.utils.cloud_utils import get_organization_default_cloud


log = BlockLogger()


def get_default_cluster_compute(
    cloud_name: Optional[str],
    project_id: Optional[str],
    api_client: Optional[DefaultApi] = None,
    anyscale_api_client: Optional[DefaultApi] = None,
) -> ComputeTemplate:
    if api_client is None:
        api_client = get_auth_api_client().api_client
    if anyscale_api_client is None:
        anyscale_api_client = get_auth_api_client().anyscale_api_client

    if cloud_name is None:
        default_cloud_name = get_organization_default_cloud(api_client)
        if default_cloud_name:
            cloud_name = default_cloud_name
        else:
            cloud_name = get_last_used_cloud(project_id, anyscale_api_client)

    cloud_id, _ = get_cloud_id_and_name(api_client, cloud_name=cloud_name)
    config_object = anyscale_api_client.get_default_compute_config(cloud_id).result
    compute_template = register_compute_template(config_object, api_client=api_client)
    return compute_template


def parse_cluster_compute_name_version(
    cluster_compute_name_version: str,
) -> Tuple[str, Optional[int]]:

    # Regex pattern to validate the cluster compute name version string:
    # - the name could not contain colon, leading or trailing spaces
    # - the version should be a positive integer
    # - the version is optional
    # - the name and version should be separated by a colon if version is provided
    COMPUTE_CONFIG_NAME_VERSION_VALIDATION_REGEX_PATTERN = (
        r"^[^\s:]+(\s+[^\s:]+)*(:[1-9]+[0-9]*)?$"
    )

    if not re.match(
        COMPUTE_CONFIG_NAME_VERSION_VALIDATION_REGEX_PATTERN,
        cluster_compute_name_version,
    ):
        raise ValueError(
            f"Invalid compute config name and version '{cluster_compute_name_version}'. Please provide name and version in format `my-compute-config:3` or provide compute configuration name `my-compute-config`."
        )

    cluster_compute_name = cluster_compute_name_version
    version = None
    if cluster_compute_name_version.count(":") == 1:
        cluster_compute_name, parsed_version = cluster_compute_name_version.split(":")
        version = int(parsed_version)

    return cluster_compute_name, version


def get_cluster_compute_from_name(
    cluster_compute_name: str,
    api_client: Optional[DefaultApi] = None,
    version: Optional[int] = None,
) -> ComputeTemplate:
    if api_client is None:
        api_client = get_auth_api_client().api_client

    enable_compute_config_versioning = api_client.check_is_feature_flag_on_api_v2_userinfo_check_is_feature_flag_on_get(
        "compute-config-versioning"
    ).result.is_on

    if enable_compute_config_versioning:
        (
            parsed_cluster_compute_name,
            parsed_version,
        ) = parse_cluster_compute_name_version(cluster_compute_name)

        if parsed_version and version:
            # TODO: after cleaning up the separate version argument, remove this check.
            raise ValueError(
                "Please provide either version in the name or as a separate argument, not both."
            )
        elif parsed_version:
            cluster_compute_name = parsed_cluster_compute_name
            version = parsed_version
        else:
            # version is not None and parsed_version is None.
            # Use non-parsed cluster_compute_name and version.
            pass

    cluster_computes = api_client.search_compute_templates_api_v2_compute_templates_search_post(
        ComputeTemplateQuery(
            orgwide=True,
            name={"equals": cluster_compute_name},
            include_anonymous=True,
            archive_status=ArchiveStatus.ALL,
            version=version,
        )
    ).results

    if len(cluster_computes) == 0:
        raise ValueError(
            f"The compute config {cluster_compute_name} does not exist or you don't have sufficient permissions."
        )
    return cluster_computes[0]


def register_compute_template(
    config_object: ComputeTemplateConfig, api_client: Optional[DefaultApi] = None,
) -> ComputeTemplate:
    """
    Register compute template with a default name and return the compute template id."""
    if api_client is None:
        api_client = get_auth_api_client().api_client
    created_template = api_client.create_compute_template_api_v2_compute_templates_post(
        create_compute_template=CreateComputeTemplate(
            config=config_object, anonymous=True,
        )
    ).result
    return created_template


def get_selected_cloud_id_or_default(
    api_client: Optional[DefaultApi] = None,
    anyscale_api_client: Optional[DefaultApi] = None,
    cluster_compute_id: Optional[str] = None,
    cluster_compute_config: Optional[ClusterComputeConfig] = None,
    cloud_id: Optional[str] = None,
    cloud_name: Optional[str] = None,
):
    """
    Gets cloud_id that is selected for the current command from the
    arguments `cloud_id`, `cloud_name`, `cluster_compute_id`, or `cluster_compute_config`.
    If the cloud_id is not selected through any of these arguments, get the default cloud
    from the default cluster compute.
    """
    api_client = api_client or get_auth_api_client(log_output=False).api_client
    anyscale_api_client = (
        anyscale_api_client or get_auth_api_client(log_output=False).anyscale_api_client
    )
    if cloud_id or cloud_name:
        parent_cloud_id, _ = get_cloud_id_and_name(
            api_client=api_client, cloud_id=cloud_id, cloud_name=cloud_name,
        )
    elif cluster_compute_id:
        parent_cloud_id = anyscale_api_client.get_cluster_compute(
            cluster_compute_id
        ).result.config.cloud_id
    elif cluster_compute_config:
        parent_cloud_id = cluster_compute_config.cloud_id
    else:
        parent_cloud_id = (
            anyscale_api_client.get_default_cluster_compute().result.config.cloud_id
        )
    return parent_cloud_id



import logging
from typing import Dict, List, Optional, Tuple
import uuid

from anyscale._private.anyscale_client.common import (
    AnyscaleClientInterface,
    WORKSPACE_CLUSTER_NAME_PREFIX,
)
from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models import (
    ComputeTemplateConfig,
    CreateInternalProductionJob,
    DecoratedComputeTemplate,
    HaJobGoalStates,
    HaJobStates,
    InternalProductionJob,
    ProductionJobStateTransition,
)
from anyscale.sdk.anyscale_client.configuration import Configuration
from anyscale.sdk.anyscale_client.models import (
    ApplyServiceModel,
    Cluster,
    ClusterCompute,
    ServiceEventCurrentState,
    ServiceModel,
)
from anyscale.utils.workspace_notification import WorkspaceNotification


block_logger = BlockLogger()
logger = logging.getLogger(__name__)

OPENAPI_NO_VALIDATION = Configuration()
OPENAPI_NO_VALIDATION.client_side_validation = False


class FakeAnyscaleClient(AnyscaleClientInterface):
    BASE_UI_URL = "http://fake.com"
    CLOUD_BUCKET = "s3://fake-bucket/{cloud_id}"
    DEFAULT_CLOUD_ID = "fake-default-cloud-id"
    DEFAULT_PROJECT_ID = "fake-default-project-id"
    DEFAULT_CLUSTER_COMPUTE_ID = "fake-default-cluster-compute-id"
    DEFAULT_CLUSTER_ENV_BUILD_ID = "fake-default-cluster-env-build-id"

    WORKSPACE_ID = "fake-workspace-id"
    WORKSPACE_CLOUD_ID = "fake-workspace-cloud-id"
    WORKSPACE_CLUSTER_ID = "fake-workspace-cluster-id"
    WORKSPACE_PROJECT_ID = "fake-workspace-project-id"
    WORKSPACE_CLUSTER_COMPUTE_ID = "fake-workspace-cluster-compute-id"
    WORKSPACE_CLUSTER_ENV_BUILD_ID = "fake-workspace-cluster-env-build-id"

    def __init__(self):
        self._image_uri_to_build_id: Dict[str, str] = {}
        self._build_id_to_image_uri: Dict[str, str] = {}
        self._containerfile_to_build_id: Dict[str, str] = {}
        self._compute_config_name_to_id: Dict[str, str] = {}
        self._compute_config_id_to_cloud_id: Dict[str, str] = {}
        self._compute_configs: Dict[str, ClusterCompute] = {}
        self._workspace_cluster: Optional[Cluster] = None
        self._workspace_dependency_tracking_enabled: bool = False
        self._services: Dict[str, ServiceModel] = {}
        self._rolled_out_model: Optional[ApplyServiceModel] = None
        self._sent_workspace_notifications: List[WorkspaceNotification] = []
        self._rolled_back_service: Optional[Tuple[str, Optional[int]]] = None
        self._terminated_service: Optional[str] = None
        self._requirements_path: Optional[str] = None
        self._upload_uri_mapping: Dict[str, str] = {}
        self._submitted_job: Optional[CreateInternalProductionJob] = None

    def get_job_ui_url(self, job_id: str) -> str:
        return f"{self.BASE_UI_URL}/jobs/{job_id}"

    def get_service_ui_url(self, service_id: str) -> str:
        return f"{self.BASE_UI_URL}/services/{service_id}"

    def set_inside_workspace(
        self,
        inside_workspace: bool,
        *,
        requirements_path: Optional[str] = None,
        cluster_name: Optional[str] = None,
    ):
        self._requirements_path = requirements_path
        if inside_workspace:
            self._workspace_cluster = Cluster(
                id=self.WORKSPACE_CLUSTER_ID,
                name=cluster_name
                if cluster_name is not None
                else WORKSPACE_CLUSTER_NAME_PREFIX + "test",
                project_id=self.WORKSPACE_PROJECT_ID,
                cluster_compute_id=self.WORKSPACE_CLUSTER_COMPUTE_ID,
                cluster_environment_build_id=self.WORKSPACE_CLUSTER_ENV_BUILD_ID,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        else:
            self._workspace_cluster = None

    def get_current_workspace_id(self) -> Optional[str]:
        return (
            self.WORKSPACE_ID
            if self.get_current_workspace_cluster() is not None
            else None
        )

    def inside_workspace(self) -> bool:
        return self.get_current_workspace_cluster() is not None

    def get_workspace_requirements_path(self) -> Optional[str]:
        return self._requirements_path

    def get_current_workspace_cluster(self) -> Optional[Cluster]:
        return self._workspace_cluster

    @property
    def sent_workspace_notifications(self) -> List[WorkspaceNotification]:
        return self._sent_workspace_notifications

    def send_workspace_notification(self, notification: WorkspaceNotification):
        if self.inside_workspace():
            self._sent_workspace_notifications.append(notification)

    def get_project_id(self) -> str:
        workspace_cluster = self.get_current_workspace_cluster()
        if workspace_cluster is not None:
            return workspace_cluster.project_id

        return self.DEFAULT_PROJECT_ID

    def set_compute_config_id_to_cloud_id_mapping(
        self, compute_config_id: str, cloud_id: str
    ):
        self._compute_config_id_to_cloud_id[compute_config_id] = cloud_id

    def get_cloud_id(
        self,
        *,
        cloud_name: Optional[str] = None,
        compute_config_id: Optional[str] = None,
    ) -> Optional[str]:
        assert not (cloud_name and compute_config_id)
        workspace_cluster = self.get_current_workspace_cluster()
        if workspace_cluster is not None:
            return self.WORKSPACE_CLOUD_ID

        if compute_config_id is not None:
            return self._compute_config_id_to_cloud_id.get(compute_config_id)

        if cloud_name is None:
            return self.DEFAULT_CLOUD_ID

        # TODO(edoakes): uncomment the below once non-default cloud is supported.
        # return f"{cloud_name}-fake-id"
        raise NotImplementedError("Only default cloud is currently implemented.")

    def add_compute_config(self, compute_config: DecoratedComputeTemplate):
        self._compute_configs[compute_config.id] = compute_config
        self._compute_config_name_to_id[compute_config.name] = compute_config.id

    def create_anonymous_compute_config(self, config: ComputeTemplateConfig) -> str:
        compute_config_id = f"compute-config-id-{str(uuid.uuid4())}"
        self.add_compute_config(
            DecoratedComputeTemplate(
                id=compute_config_id,
                name=f"compute-config-{str(uuid.uuid4())}",
                config=config,
                anonymous=True,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
        )
        return compute_config_id

    def get_compute_config(
        self, compute_config_id: str
    ) -> Optional[DecoratedComputeTemplate]:
        return self._compute_configs.get(compute_config_id, None)

    def get_compute_config_id(
        self, compute_config_name: Optional[str] = None
    ) -> Optional[str]:
        if compute_config_name is not None:
            return self._compute_config_name_to_id.get(compute_config_name)

        workspace_cluster = self.get_current_workspace_cluster()
        if workspace_cluster is not None:
            return workspace_cluster.cluster_compute_id

        return self.DEFAULT_CLUSTER_COMPUTE_ID

    def set_image_uri_mapping(self, image_uri: str, build_id: str):
        self._image_uri_to_build_id[image_uri] = build_id
        self._build_id_to_image_uri[build_id] = image_uri

    def set_containerfile_mapping(self, containerfile: str, build_id: str):
        self._containerfile_to_build_id[containerfile] = build_id

    def get_default_build_id(self) -> str:
        workspace_cluster = self.get_current_workspace_cluster()
        if workspace_cluster is not None:
            return workspace_cluster.cluster_environment_build_id

        return self.DEFAULT_CLUSTER_ENV_BUILD_ID

    def get_cluster_env_build_id_from_containerfile(
        self, cluster_env_name: str, containerfile: str  # noqa: ARG002
    ) -> str:
        if containerfile in self._containerfile_to_build_id:
            return self._containerfile_to_build_id[containerfile]
        else:
            raise ValueError(f"The containerfile '{containerfile}' does not exist.")

    def get_cluster_env_build_id(self, image_uri: str) -> str:
        if image_uri in self._image_uri_to_build_id:
            return self._image_uri_to_build_id[image_uri]
        else:
            raise ValueError(f"The image_uri '{image_uri}' does not exist.")

    def get_cluster_env_build_image_uri(
        self, cluster_env_build_id: str
    ) -> Optional[str]:
        return self._build_id_to_image_uri.get(cluster_env_build_id, None)

    def update_service(self, model: ServiceModel):
        self._services[model.id] = model

    def get_service(self, name: str) -> Optional[ServiceModel]:
        for service in self._services.values():
            if service.name == name:
                return service

        return None

    @property
    def rolled_out_model(self) -> Optional[ApplyServiceModel]:
        return self._rolled_out_model

    def rollout_service(self, model: ApplyServiceModel) -> ServiceModel:
        self._rolled_out_model = model
        existing_service = self.get_service(model.name)
        if existing_service is not None:
            service_id = existing_service.id
        else:
            service_id = f"service-id-{str(uuid.uuid4())}"

        service = ServiceModel(
            id=service_id,
            name=model.name,
            current_state=ServiceEventCurrentState.RUNNING,
            base_url="http://fake-service-url",
            auth_token="fake-auth-token",
            local_vars_configuration=OPENAPI_NO_VALIDATION,
        )

        self.update_service(service)
        return service

    @property
    def rolled_back_service(self) -> Optional[Tuple[str, Optional[int]]]:
        return self._rolled_back_service

    def rollback_service(
        self, service_id: str, *, max_surge_percent: Optional[int] = None
    ):
        self._rolled_back_service = (service_id, max_surge_percent)

    @property
    def terminated_service(self) -> Optional[str]:
        return self._terminated_service

    def terminate_service(self, service_id: str):
        self._terminated_service = service_id

    @property
    def submitted_job(self) -> Optional[CreateInternalProductionJob]:
        return self._submitted_job

    def submit_job(self, model: CreateInternalProductionJob) -> InternalProductionJob:
        self._submitted_job = model

        return InternalProductionJob(
            id=f"job-{str(uuid.uuid4())}",
            name=model.name,
            config=model.config,
            state=ProductionJobStateTransition(
                current_state=HaJobStates.PENDING,
                goal_state=HaJobGoalStates.SUCCESS,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
            local_vars_configuration=OPENAPI_NO_VALIDATION,
        )

    def upload_local_dir_to_cloud_storage(
        self,
        local_dir: str,  # noqa: ARG002
        *,
        cloud_id: str,
        excludes: Optional[List[str]] = None,  # noqa: ARG002
    ) -> str:
        # Ensure that URIs are consistent for the same passed directory.
        if local_dir in self._upload_uri_mapping:
            uri = self._upload_uri_mapping[local_dir]
        else:
            uri = str(uuid.uuid4())
            self._upload_uri_mapping[local_dir] = uri

        bucket = self.CLOUD_BUCKET.format(cloud_id=cloud_id)
        return f"{bucket}/fake_pkg_{uri}.zip"



import copy
from typing import Optional
import uuid

from anyscale._private.sdk.workload_sdk import WorkloadSDK
from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models import (
    CreateInternalProductionJob,
    InternalProductionJob,
    ProductionJobConfig,
)
from anyscale.job.models import JobConfig


logger = BlockLogger()


class JobSDK(WorkloadSDK):
    def _override_runtime_env(
        self,
        config: JobConfig,
        *,
        autopopulate_in_workspace: bool = True,
        cloud_id: Optional[str] = None,
    ) -> JobConfig:
        """Overrides the runtime_env in the config.

        Local directories specified in the 'working_dir' or 'py_modules' fields will be
        uploaded and replaced with the resulting remote URIs.

        Requirements files will be loaded and populated into the 'pip' field.

        If autopopulate_from_workspace is passed and this code is running inside a
        workspace, the following defaults will be applied:
            - 'working_dir' will be set to '.'.
            - 'pip' will be set to the workspace-managed requirements file.
        """
        new_runtime_env = (
            copy.deepcopy(config.runtime_env) if config.runtime_env else {}
        )
        [new_runtime_env] = self.override_and_upload_local_dirs(
            [new_runtime_env],
            working_dir_override=config.working_dir,
            excludes_override=config.excludes,
            cloud_id=cloud_id,
            autopopulate_in_workspace=autopopulate_in_workspace,
        )
        [new_runtime_env] = self.override_and_load_requirements_files(
            [new_runtime_env],
            requirements_override=config.requirements,
            autopopulate_in_workspace=autopopulate_in_workspace,
        )

        if new_runtime_env:
            # If `None` or `{}` was passed in and no overrides happened, leave it as-is.
            config = config.options(runtime_env=new_runtime_env)

        return config.options(requirements=None, working_dir=None, excludes=None,)

    def _get_default_name(self) -> str:
        """Get a default name for the job.

        If running inside a workspace, this is generated from the workspace name,
        else it generates a random name.
        """
        # TODO(edoakes): generate two random words instead of UUID here.
        name = f"job-{self.get_current_workspace_name() or str(uuid.uuid4())}"
        self.logger.info(f"No name was specified, using default: '{name}'.")
        return name

    def submit(self, config: JobConfig) -> str:
        name = config.name or self._get_default_name()
        if config.containerfile is not None:
            build_id = self.client.get_cluster_env_build_id_from_containerfile(
                cluster_env_name=f"image-for-{name}",
                containerfile=self.get_containerfile_contents(config.containerfile),
            )
        elif config.image_uri is not None:
            build_id = self.client.get_cluster_env_build_id(image_uri=config.image_uri,)
        else:
            build_id = self.client.get_default_build_id()

        compute_config_id = self.resolve_compute_config_to_id(config.compute_config)

        # If a compute config was specified, we need to make sure to used the correct
        # cloud_id when uploading local directories.
        cloud_id = self.client.get_cloud_id(
            compute_config_id=None
            if config.compute_config is None
            else compute_config_id
        )
        config = self._override_runtime_env(config, cloud_id=cloud_id)

        job: InternalProductionJob = self.client.submit_job(
            CreateInternalProductionJob(
                name=name,
                # TODO(edoakes): should we support a description?
                description=None,
                project_id=self.client.get_project_id(),
                workspace_id=self.client.get_current_workspace_id(),
                config=ProductionJobConfig(
                    entrypoint=config.entrypoint,
                    runtime_env=config.runtime_env,
                    build_id=build_id,
                    compute_config_id=compute_config_id,
                    max_retries=config.max_retries,
                ),
                # TODO(edoakes): support job queue config.
                job_queue_config=None,
            )
        )

        self.logger.info(f"Job '{name}' submitted, ID: '{job.id}'.")
        self.logger.info(
            f"View the job in the UI: {self.client.get_job_ui_url(job.id)}"
        )
        return job.id



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional

from anyscale._private.models import ModelBase, WorkloadConfig


@dataclass(frozen=True)
class ServiceConfig(WorkloadConfig):
    applications: List[Dict[str, Any]] = field(default_factory=list, repr=False)

    def _validate_applications(self, applications: List[Dict[str, Any]]):
        if not isinstance(applications, list):
            raise TypeError("'applications' must be a list.")
        if len(applications) == 0:
            raise ValueError("'applications' cannot be empty.")

        # Validate import paths.
        for app in applications:
            import_path = app.get("import_path", None)
            if not import_path:
                raise ValueError("Every application must specify an import path.")

            if not isinstance(import_path, str):
                raise TypeError(f"'import_path' must be a string, got: {import_path}")

            if (
                import_path.count(":") != 1
                or import_path.rfind(":") in {0, len(import_path) - 1}
                or import_path.rfind(".") in {0, len(import_path) - 1}
            ):
                raise ValueError(
                    f"'import_path' must be of the form: 'module.optional_submodule:app', but got: '{import_path}'."
                )

    query_auth_token_enabled: bool = field(default=True, repr=False)

    def _validate_query_auth_token_enabled(self, query_auth_token_enabled: bool):
        if not isinstance(query_auth_token_enabled, bool):
            raise TypeError("'query_auth_token_enabled' must be a boolean.")

    grpc_options: Optional[Dict[str, Any]] = field(default=None, repr=False)

    def _validate_grpc_options(self, grpc_options: Optional[Dict[str, Any]]):
        """Validate the `grpc_options` field.

        This will be passed through as part of the Ray Serve config, but some fields are
        disallowed (not valid when deploying Anyscale services).
        """
        if grpc_options is None:
            return
        elif not isinstance(grpc_options, dict):
            raise TypeError("'grpc_options' must be a dict.")

        banned_options = {
            "port",
        }
        banned_options_passed = {o for o in banned_options if o in grpc_options}
        if len(banned_options_passed) > 0:
            raise ValueError(
                "The following provided 'grpc_options' are not permitted "
                f"in Anyscale: {banned_options_passed}."
            )

    http_options: Optional[Dict[str, Any]] = field(default=None, repr=False)

    def _validate_http_options(self, http_options: Optional[Dict[str, Any]]):
        """Validate the `http_options` field.

        This will be passed through as part of the Ray Serve config, but some fields are
        disallowed (not valid when deploying Anyscale services).
        """
        if http_options is None:
            return
        elif not isinstance(http_options, dict):
            raise TypeError("'http_options' must be a dict.")

        banned_options = {"host", "port", "root_path"}
        banned_options_passed = {o for o in banned_options if o in http_options}
        if len(banned_options_passed) > 0:
            raise ValueError(
                "The following provided 'http_options' are not permitted "
                f"in Anyscale: {banned_options_passed}."
            )

    logging_config: Optional[Dict[str, Any]] = field(default=None, repr=False)

    def _validate_logging_config(self, logging_config: Optional[Dict[str, Any]]):
        if logging_config is not None and not isinstance(logging_config, dict):
            raise TypeError("'logging_config' must be a dict.")

    ray_gcs_external_storage_config: Optional[Dict[str, Any]] = field(
        default=None, repr=False
    )

    def _validate_ray_gcs_external_storage_config(
        self, ray_gcs_external_storage_config: Optional[Dict[str, Any]]
    ):
        if ray_gcs_external_storage_config is not None and not isinstance(
            ray_gcs_external_storage_config, dict
        ):
            raise TypeError("'ray_gcs_external_storage_config' must be a dict.")


class ServiceState(str, Enum):
    UNKNOWN = "UNKNOWN"
    STARTING = "STARTING"
    RUNNING = "RUNNING"
    # TODO(edoakes): UPDATING comes up while rolling out and rolling back.
    # This is very unexpected from a customer's point of view, we should fix it.
    UPDATING = "UPDATING"
    ROLLING_OUT = "ROLLING_OUT"
    ROLLING_BACK = "ROLLING_BACK"
    TERMINATING = "TERMINATING"
    TERMINATED = "TERMINATED"
    UNHEALTHY = "UNHEALTHY"
    SYSTEM_FAILURE = "SYSTEM_FAILURE"

    def __str__(self):
        return self.name


# TODO(edoakes): we should have a corresponding ServiceVersionState.
@dataclass(frozen=True)
class ServiceVersionStatus(ModelBase):
    name: str
    weight: int
    config: ServiceConfig = field(repr=False)

    def _validate_name(self, name: str):
        if not isinstance(name, str):
            raise TypeError("'name' must be a string.")

    def _validate_weight(self, weight: int):
        if not isinstance(weight, int):
            raise TypeError("'weight' must be an int.")

    def _validate_config(self, config: ServiceConfig):
        if not isinstance(config, ServiceConfig):
            raise TypeError("'config' must be a ServiceConfig.")


@dataclass(frozen=True)
class ServiceStatus(ModelBase):
    id: str

    def _validate_id(self, id: str):  # noqa: A002
        if not isinstance(id, str):
            raise TypeError("'id' must be a string.")

    name: str

    def _validate_name(self, name: str):
        if not isinstance(name, str):
            raise TypeError("'name' must be a string.")

    state: ServiceState

    def _validate_state(self, state: ServiceState):
        if isinstance(state, str):
            # This will raise a ValueError if the state is unrecognized.
            object.__setattr__(self, "state", ServiceState(state))
        elif not isinstance(state, ServiceState):
            raise TypeError("'state' must be a ServiceState.")

    query_url: str = field(repr=False)

    def _validate_query_url(self, query_url: str):
        if not isinstance(query_url, str):
            raise TypeError("'query_url' must be a string.")

    query_auth_token: Optional[str] = field(default=None, repr=False)

    def _validate_query_auth_token(self, query_auth_token: Optional[str]):
        if query_auth_token is not None and not isinstance(query_auth_token, str):
            raise TypeError("'query_auth_token' must be a string.")

    primary_version: Optional[ServiceVersionStatus] = field(default=None, repr=False)

    def _validate_primary_version(
        self, primary_version: Optional[ServiceVersionStatus]
    ):
        if primary_version is not None and not isinstance(
            primary_version, ServiceVersionStatus
        ):
            raise TypeError("'primary_version' must be a ServiceVersionStatus.")

    canary_version: Optional[ServiceVersionStatus] = field(default=None, repr=False)

    def _validate_canary_version(self, canary_version: Optional[ServiceVersionStatus]):
        if canary_version is not None and not isinstance(
            canary_version, ServiceVersionStatus
        ):
            raise TypeError("'canary_version' must be a ServiceVersionStatus.")



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class Card(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'id': 'CardId',
        'dismissal_type': 'list[DismissalType]',
        'visibility': 'Visibility'
    }

    attribute_map = {
        'id': 'id',
        'dismissal_type': 'dismissal_type',
        'visibility': 'visibility'
    }

    def __init__(self, id=None, dismissal_type=None, visibility=None, local_vars_configuration=None):  # noqa: E501
        """Card - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._id = None
        self._dismissal_type = None
        self._visibility = None
        self.discriminator = None

        self.id = id
        self.dismissal_type = dismissal_type
        self.visibility = visibility

    @property
    def id(self):
        """Gets the id of this Card.  # noqa: E501

        Unique card id.  # noqa: E501

        :return: The id of this Card.  # noqa: E501
        :rtype: CardId
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this Card.

        Unique card id.  # noqa: E501

        :param id: The id of this Card.  # noqa: E501
        :type: CardId
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def dismissal_type(self):
        """Gets the dismissal_type of this Card.  # noqa: E501

        Ways the card can be dismissed. If dismissal_type={DismissalType.manual, DismissalType.automatic}, the card will be visible until it is dismissed either manually or automatically.  # noqa: E501

        :return: The dismissal_type of this Card.  # noqa: E501
        :rtype: list[DismissalType]
        """
        return self._dismissal_type

    @dismissal_type.setter
    def dismissal_type(self, dismissal_type):
        """Sets the dismissal_type of this Card.

        Ways the card can be dismissed. If dismissal_type={DismissalType.manual, DismissalType.automatic}, the card will be visible until it is dismissed either manually or automatically.  # noqa: E501

        :param dismissal_type: The dismissal_type of this Card.  # noqa: E501
        :type: list[DismissalType]
        """
        if self.local_vars_configuration.client_side_validation and dismissal_type is None:  # noqa: E501
            raise ValueError("Invalid value for `dismissal_type`, must not be `None`")  # noqa: E501

        self._dismissal_type = dismissal_type

    @property
    def visibility(self):
        """Gets the visibility of this Card.  # noqa: E501

        Group of users the card is visible to.  # noqa: E501

        :return: The visibility of this Card.  # noqa: E501
        :rtype: Visibility
        """
        return self._visibility

    @visibility.setter
    def visibility(self, visibility):
        """Sets the visibility of this Card.

        Group of users the card is visible to.  # noqa: E501

        :param visibility: The visibility of this Card.  # noqa: E501
        :type: Visibility
        """
        if self.local_vars_configuration.client_side_validation and visibility is None:  # noqa: E501
            raise ValueError("Invalid value for `visibility`, must not be `None`")  # noqa: E501

        self._visibility = visibility

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Card):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, Card):
            return True

        return self.to_dict() != other.to_dict()



import dataclasses
import enum
import ipaddress
import socket
from typing import List, Optional, Set, Union

from anyscale.cli_logger import CloudSetupLogger
from anyscale.client.openapi_client.models import CloudAnalyticsEventCloudResource
from anyscale.utils.cloud_utils import CloudSetupError


_SOCKET_TO_PROTO_MAP = {
    getattr(socket, attr): attr.replace("IPPROTO_", "").lower()
    for attr in dir(socket)
    if attr.startswith("IPPROTO_")
}

IPNetwork = Union[ipaddress.IPv4Network, ipaddress.IPv6Network]


@dataclasses.dataclass
class CapacityThreshold:
    min_network_size: IPNetwork
    warn_network_size: IPNetwork
    resource_type: str
    cloud_resource: CloudAnalyticsEventCloudResource

    def verify_network_capacity(
        self, *, cidr_block_str: str, resource_name: str, logger: CloudSetupLogger
    ) -> bool:
        cidr_block = ipaddress.ip_network(cidr_block_str, strict=False)

        min_hosts = self.min_network_size.num_addresses
        warn_hosts = self.warn_network_size.num_addresses

        if cidr_block.num_addresses < min_hosts:
            logger.error(
                f"The provided {self.resource_type} ({resource_name})'s CIDR block ({cidr_block}) is too"
                f" small. We want at least {min_hosts} addresses,"
                f" but this {self.resource_type} only has {cidr_block.num_addresses}. Please reach out to"
                f" support if this is an issue!"
            )
            logger.log_resource_error(
                self.cloud_resource, CloudSetupError.CIDR_BLOCK_TOO_SMALL
            )
            return False
        elif cidr_block.num_addresses < warn_hosts:
            logger.warning(
                f"The provided {self.resource_type} ({resource_name})'s CIDR block ({cidr_block}) is probably"
                f" too small. We suggest at least {warn_hosts}"
                f" addresses, but this {self.resource_type} only supports up to"
                f" {cidr_block.num_addresses} addresses."
            )

        return True


AWS_VPC_CAPACITY = CapacityThreshold(
    min_network_size=ipaddress.ip_network("10.0.0.0/24"),
    warn_network_size=ipaddress.ip_network("10.0.0.0/20"),
    resource_type="VPC",
    cloud_resource=CloudAnalyticsEventCloudResource.AWS_VPC,
)

AWS_SUBNET_CAPACITY = CapacityThreshold(
    min_network_size=ipaddress.ip_network("10.0.0.0/28"),
    warn_network_size=ipaddress.ip_network("10.0.0.0/24"),
    resource_type="Subnet",
    cloud_resource=CloudAnalyticsEventCloudResource.AWS_SUBNET,
)

GCP_SUBNET_CAPACITY = dataclasses.replace(
    AWS_VPC_CAPACITY,
    resource_type="Subnet",
    cloud_resource=CloudAnalyticsEventCloudResource.GCP_SUBNET,
)


class Direction(enum.Enum):
    INGRESS = "INGRESS"
    EGRESS = "EGRESS"


class Protocol(enum.Enum):
    all = "all"
    tcp = "tcp"
    udp = "udp"
    icmp = "icmp"
    other = "other"

    @classmethod
    def from_val(cls, val: Union[str, int]):
        # Convert protocol numbers to strings
        if isinstance(val, int) or val.isdigit():
            val = _SOCKET_TO_PROTO_MAP.get(int(val), "other")

        if hasattr(cls, val.lower()):  # type: ignore
            return cls(val.lower())  # type: ignore

        return Protocol.other


@dataclasses.dataclass
class FirewallRule:
    direction: Direction
    protocol: Protocol
    network: IPNetwork
    ports: Optional[Set[int]]

    def __repr__(self):
        return f"FirewallRule(direction='{self.direction.value}', protocol='{self.protocol.value}', network='{self.network}', num_ports={len(self.ports)})"

    def _canonicalize_ports(self):
        if self.ports is None or len(self.ports) == 0:
            self.ports = set(range(1, 65536))
            return
        ports = self.ports
        self.ports = set()
        for port in ports:
            if "-" in str(port):
                lower, upper = port.split("-")
                self.ports.update(range(int(lower), int(upper) + 1))
            else:
                self.ports.add(int(port))

    def __post_init__(self):
        self.direction = Direction(self.direction)

        self._canonicalize_ports()


def check_inbound_firewall_permissions(
    rules: List[FirewallRule],
    protocol: Protocol,
    ports: Optional[Set[int]],
    source_range: IPNetwork,
    check_ingress_source_range: bool = True,
) -> bool:
    ports_remaining = ports.copy() if ports is not None else set(range(1, 65536))
    for rule in rules:

        if (
            rule.direction != Direction.INGRESS
            or not isinstance(source_range, type(rule.network))
            or not source_range.subnet_of(rule.network)  # type: ignore
        ) and check_ingress_source_range:
            continue

        if rule.protocol == Protocol.all or protocol == rule.protocol:
            ports_remaining.difference_update(rule.ports or {})

    # Port 0 is not a thing, so just account for it being weird
    return len(ports_remaining) == 0 or ports_remaining == {0}



from typing import Optional

from anyscale.api_utils.exceptions.job_errors import NoJobRunError
from anyscale.sdk.anyscale_client.api.default_api import DefaultApi as BaseApi
from anyscale.sdk.anyscale_client.models.production_job import ProductionJob


def _get_job_run_id(
    base_api: BaseApi,
    *,
    job_id: Optional[str] = None,
    job_run_id: Optional[str] = None,
) -> str:
    assert bool(job_id) != bool(
        job_run_id
    ), "Exactly one of `job_id` or `job_run_id` must be provided."
    if job_id:
        prod_job: ProductionJob = base_api.get_production_job(job_id).result
        if not prod_job.last_job_run_id:
            raise NoJobRunError(f"Production job {job_id} has no job runs.")
        job_run_id = prod_job.last_job_run_id
    return job_run_id  # type: ignore



import logging
import os
import pathlib
import re
import time
from typing import Callable, Dict, List, Optional, Tuple

import requests

from anyscale._private.anyscale_client.common import (
    AnyscaleClientInterface,
    DEFAULT_PYTHON_VERSION,
    DEFAULT_RAY_VERSION,
    RUNTIME_ENV_PACKAGE_FORMAT,
)
from anyscale.authenticate import AuthenticationBlock, get_auth_api_client
from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.api.default_api import DefaultApi as InternalApi
from anyscale.client.openapi_client.models import (
    ArchiveStatus,
    CloudDataBucketFileType,
    CloudDataBucketPresignedUploadInfo,
    CloudDataBucketPresignedUploadRequest,
    ComputeTemplateConfig,
    ComputeTemplateQuery,
    CreateComputeTemplate,
    CreateInternalProductionJob,
    DecoratedComputeTemplate,
    InternalProductionJob,
)
from anyscale.cluster_compute import parse_cluster_compute_name_version
from anyscale.sdk.anyscale_client.api.default_api import DefaultApi as ExternalApi
from anyscale.sdk.anyscale_client.models import (
    ApplyServiceModel,
    Cloud,
    Cluster,
    ClusterCompute,
    ClusterComputeConfig,
    ClusterEnvironment,
    ClusterEnvironmentBuild,
    ClusterEnvironmentBuildStatus,
    Project,
    RollbackServiceModel,
    ServiceModel,
)
from anyscale.sdk.anyscale_client.models.create_cluster_environment import (
    CreateClusterEnvironment,
)
from anyscale.sdk.anyscale_client.models.create_cluster_environment_build import (
    CreateClusterEnvironmentBuild,
)
from anyscale.sdk.anyscale_client.rest import ApiException
from anyscale.util import (
    get_cluster_model_for_current_workspace,
    get_endpoint,
    is_anyscale_workspace,
)
from anyscale.utils.runtime_env import (
    is_workspace_dependency_tracking_disabled,
    WORKSPACE_REQUIREMENTS_FILE_PATH,
    zip_local_dir,
)
from anyscale.utils.workspace_notification import (
    WORKSPACE_NOTIFICATION_ADDRESS,
    WorkspaceNotification,
)


block_logger = BlockLogger()
logger = logging.getLogger(__name__)

WORKSPACE_ID_ENV_VAR = "ANYSCALE_EXPERIMENTAL_WORKSPACE_ID"


class AnyscaleClient(AnyscaleClientInterface):
    # Number of entries to fetch per request for list endpoints.
    LIST_ENDPOINT_COUNT = 50

    def __init__(
        self,
        *,
        api_clients: Optional[Tuple[ExternalApi, InternalApi]] = None,
        sleep: Optional[Callable[[float], None]] = None,
        workspace_requirements_file_path: str = WORKSPACE_REQUIREMENTS_FILE_PATH,
    ):
        if api_clients is None:
            auth_block: AuthenticationBlock = get_auth_api_client(
                raise_structured_exception=True
            )
            api_clients = (auth_block.anyscale_api_client, auth_block.api_client)

        self._external_api_client, self._internal_api_client = api_clients
        self._workspace_requirements_file_path = workspace_requirements_file_path
        self._sleep = sleep or time.sleep

        # Cached IDs and models to avoid duplicate lookups.
        self._default_project_id: Optional[str] = None
        self._cloud_id_cache: Dict[Optional[str], str] = {}
        self._current_workspace_cluster: Optional[Cluster] = None

    def get_job_ui_url(self, job_id: str) -> str:
        return get_endpoint(f"/jobs/{job_id}")

    def get_service_ui_url(self, service_id: str) -> str:
        return get_endpoint(f"/services/{service_id}")

    def get_current_workspace_id(self) -> Optional[str]:
        return os.environ.get(WORKSPACE_ID_ENV_VAR, None)

    def inside_workspace(self) -> bool:
        return self.get_current_workspace_id() is not None

    def get_workspace_requirements_path(self) -> Optional[str]:
        if (
            not self.inside_workspace()
            or is_workspace_dependency_tracking_disabled()
            or not pathlib.Path(self._workspace_requirements_file_path).is_file()
        ):
            return None

        return self._workspace_requirements_file_path

    def get_current_workspace_cluster(self) -> Optional[Cluster]:
        # Checks for the existence of the ANYSCALE_EXPERIMENTAL_WORKSPACE_ID env var.
        if not is_anyscale_workspace():
            return None

        if self._current_workspace_cluster is None:
            # Picks up the cluster ID from the ANYSCALE_SESSION_ID env var.
            self._current_workspace_cluster = get_cluster_model_for_current_workspace(
                self._external_api_client
            )

        return self._current_workspace_cluster

    def get_project_id(self) -> str:
        workspace_cluster = self.get_current_workspace_cluster()
        if workspace_cluster is not None:
            return workspace_cluster.project_id

        if self._default_project_id is None:
            default_project: Project = self._external_api_client.get_default_project().result
            self._default_project_id = default_project.id

        return self._default_project_id

    def _get_cloud_id_for_compute_config_id(self, compute_config_id: str) -> str:
        cluster_compute: ClusterCompute = self._external_api_client.get_cluster_compute(
            compute_config_id
        ).result
        cluster_compute_config: ClusterComputeConfig = cluster_compute.config
        return cluster_compute_config.cloud_id

    def get_cloud_id(
        self, cloud_name: Optional[str] = None, compute_config_id: Optional[str] = None
    ) -> Optional[str]:
        if cloud_name is not None:
            raise NotImplementedError("Only default cloud is currently implemented.")

        if compute_config_id is not None:
            return self._get_cloud_id_for_compute_config_id(compute_config_id)

        if cloud_name in self._cloud_id_cache:
            return self._cloud_id_cache[cloud_name]

        workspace_cluster = self.get_current_workspace_cluster()
        if workspace_cluster is not None:
            # NOTE(edoakes): the Cluster model has a compute_config_config model that includes
            # its cloud ID, but it's not always populated.
            # TODO(edoakes): add cloud_id to the Cluster model to avoid a second RTT.
            cloud_id = self._get_cloud_id_for_compute_config_id(
                workspace_cluster.cluster_compute_id
            )
        else:
            result: Cloud = self._external_api_client.get_default_cloud().result
            cloud_id = result.id

        self._cloud_id_cache[cloud_name] = cloud_id
        return cloud_id

    def create_anonymous_compute_config(self, config: ComputeTemplateConfig) -> str:
        return self._internal_api_client.create_compute_template_api_v2_compute_templates_post(
            create_compute_template=CreateComputeTemplate(
                config=config, anonymous=True,
            )
        ).result.id

    def get_compute_config(
        self, compute_config_id: str
    ) -> Optional[DecoratedComputeTemplate]:
        try:
            cluster_compute: ClusterCompute = self._internal_api_client.get_compute_template_api_v2_compute_templates_template_id_get(
                compute_config_id
            ).result
            return cluster_compute
        except ApiException as e:
            if e.status == 404:
                return None

            raise e from None

    def get_compute_config_id(
        self, compute_config_name: Optional[str] = None
    ) -> Optional[str]:
        if compute_config_name is not None:
            name, version = parse_cluster_compute_name_version(compute_config_name)
            if version is None:
                # Setting `version=-1` will return only the latest version if there are multiple.
                version = -1
            cluster_computes = self._internal_api_client.search_compute_templates_api_v2_compute_templates_search_post(
                ComputeTemplateQuery(
                    orgwide=True,
                    name={"equals": name},
                    include_anonymous=True,
                    archive_status=ArchiveStatus.NOT_ARCHIVED,
                    version=version,
                )
            ).results

            if len(cluster_computes) == 0:
                return None

            compute_template: DecoratedComputeTemplate = cluster_computes[0]
            return compute_template.id

        # If the compute config name is not provided, we pick an appropriate default.
        #
        #   - If running in a workspace:
        #       * If auto_select_worker_config enabled: we switch over to a standardized
        #         default compute config (copying over any cluster-level attributes, e.g.
        #         max-gpus).
        #       * Otherwise, we use the workspace's compute config.
        #
        #   - Otherwise, we use the default compute config provided by the API.

        workspace_cluster = self.get_current_workspace_cluster()
        if workspace_cluster is not None:
            workspace_compute_config: DecoratedComputeTemplate = self.get_compute_config(
                workspace_cluster.cluster_compute_id
            )
            workspace_config: ClusterComputeConfig = workspace_compute_config.config
            if workspace_config.auto_select_worker_config:
                standard_template = self._build_standard_compute_template_from_existing_auto_config(
                    workspace_config
                )
                return self.create_anonymous_compute_config(standard_template)
            else:
                return workspace_cluster.cluster_compute_id

        # NOTE(edoakes): for some reason the API endpoint is called "cluster compute" but this
        # is the non-deprecated API to use for compute configs.
        return self._external_api_client.get_default_cluster_compute().result.id

    def _build_standard_compute_template_from_existing_auto_config(
        self, compute_config: ClusterComputeConfig
    ) -> ComputeTemplateConfig:
        """
        Build a standard compute template config from an existing compute config.

        1. Pull the default compute template config.
        2. Disable scheduling on the head node.
        3. Enable auto_select_worker_config.
        4. Copy any cluster-level flags from the provided compute config to the template.
        """
        # Retrieve the default cluster compute config for the cloud.
        default_compute_template: DecoratedComputeTemplate = self._external_api_client.get_default_cluster_compute(
            cloud_id=compute_config.cloud_id,
        ).result.config

        # Disable head node scheduling.
        if default_compute_template.head_node_type.resources is None:
            default_compute_template.head_node_type.resources = {}
        default_compute_template.head_node_type.resources["CPU"] = 0

        # Ensure auto_select_worker_config is enabled.
        default_compute_template.auto_select_worker_config = True

        # Copy flags set at a cluster level over to the workspace.
        #
        # NOTE (shomilj): If there are more attributes we want to
        # persist from the provided compute config --> the compute
        # config used for deploying the service, we should copy them
        # over here.
        default_compute_template.flags = compute_config.flags

        return default_compute_template

    def get_cluster_env_build_image_uri(
        self, cluster_env_build_id: str
    ) -> Optional[str]:
        try:
            build: ClusterEnvironmentBuild = self._external_api_client.get_cluster_environment_build(
                cluster_env_build_id
            ).result
            return build.docker_image_name
        except ApiException as e:
            if e.status == 404:
                return None

            raise e from None

    def get_default_build_id(self) -> str:
        """Get default build id.

        If running in a workspace, it will return the workspace's cluster environment build ID.
        Else it will return the default cluster environment build ID.
        """
        workspace_cluster = self.get_current_workspace_cluster()
        if workspace_cluster is not None:
            return workspace_cluster.cluster_environment_build_id
        result: ClusterEnvironmentBuild = self._external_api_client.get_default_cluster_environment_build(
            DEFAULT_PYTHON_VERSION, DEFAULT_RAY_VERSION,
        ).result
        return result.id

    @staticmethod
    def _get_cluster_env_name_from_image_uri(image_uri: str) -> str:
        if len(image_uri) == 0:
            raise ValueError("image_uri cannot be empty.")

        pattern = re.compile("^[A-Za-z0-9_-]+$")
        # Keep only characters that match the pattern
        escaped = []
        for c in image_uri:
            if not pattern.match(c):
                escaped.append("-")
            else:
                escaped.append(c)
        return "".join(escaped)

    def _get_cluster_env_by_name(self, name: str) -> Optional[ClusterEnvironment]:
        cluster_envs = self._external_api_client.search_cluster_environments(
            {
                "name": {"equals": name},
                "paging": {"count": 1},
                "include_anonymous": True,
            }
        ).results
        return cluster_envs[0] if cluster_envs else None

    def _wait_for_build_to_succeed(
        self, build_id: str, poll_interval_seconds=3, timeout_secs=3600,
    ):
        """Periodically check the status of the build operation until it completes.
        Raise a RuntimeError if the build fails or cancelled.
        Raise a TimeoutError if the build does not complete within the timeout.
        """
        elapsed_secs = 0
        block_logger.info(f"Waiting for image build ({build_id}) to complete.")
        while elapsed_secs < timeout_secs:
            build = self._external_api_client.get_cluster_environment_build(
                build_id
            ).result
            if build.status == ClusterEnvironmentBuildStatus.SUCCEEDED:
                return
            elif build.status == ClusterEnvironmentBuildStatus.FAILED:
                raise RuntimeError(f"Image build {build_id} failed.")
            elif build.status == ClusterEnvironmentBuildStatus.CANCELED:
                raise RuntimeError(f"Image build {build_id} unexpectedly cancelled.")

            elapsed_secs += poll_interval_seconds
            block_logger.info(
                f"Waiting for image build to complete. Elapsed time: {elapsed_secs} seconds",
            )
            self._sleep(poll_interval_seconds)
        raise TimeoutError(
            f"Timed out waiting for image build {build_id} to complete after {timeout_secs}s."
        )

    def _find_or_create_cluster_env(self, cluster_env_name: str) -> ClusterEnvironment:
        existing_cluster_env = self._get_cluster_env_by_name(cluster_env_name)
        if existing_cluster_env is not None:
            return existing_cluster_env

        # this creates a cluster env only and it does not trigger a build to avoid the race condition
        # when creating a cluster environment and a build at the same time and then list builds for the cluster environment later.
        # ```
        #     cluster_env == self._external_api_client.create_cluster_environment(..., image_uri=image_uri)
        #     build == self._external_api_client.create_cluster_environment_build(..., cluster_env_id=cluster_env.id)
        # ```
        # The race condition can happen if another build for the same cluster envrionment is created between the two calls.
        cluster_environment = self._external_api_client.create_cluster_environment(
            CreateClusterEnvironment(name=cluster_env_name, anonymous=True)
        ).result
        return cluster_environment

    def get_cluster_env_build_id_from_containerfile(
        self, cluster_env_name: str, containerfile: str
    ) -> str:
        cluster_env = self._find_or_create_cluster_env(cluster_env_name)
        cluster_env_builds = self._external_api_client.list_cluster_environment_builds(
            cluster_environment_id=cluster_env.id, count=self.LIST_ENDPOINT_COUNT
        ).results
        for build in cluster_env_builds:
            if (
                build.status == ClusterEnvironmentBuildStatus.SUCCEEDED
                and build.containerfile == containerfile
            ):
                return build.id

        build_op = self._external_api_client.create_cluster_environment_build(
            CreateClusterEnvironmentBuild(
                cluster_environment_id=cluster_env.id, containerfile=containerfile,
            )
        ).result

        self._wait_for_build_to_succeed(build_op.cluster_environment_build_id)

        return build_op.cluster_environment_build_id

    def get_cluster_env_build_id(self, image_uri: str) -> str:
        cluster_env_name = self._get_cluster_env_name_from_image_uri(image_uri)
        cluster_env = self._find_or_create_cluster_env(cluster_env_name)

        # since we encode the image_uri into the cluster env name, there should exist one and only one build that matches the image_uri.
        cluster_env_builds = self._external_api_client.list_cluster_environment_builds(
            cluster_environment_id=cluster_env.id, count=1
        ).results
        build = cluster_env_builds[0] if cluster_env_builds else None
        if (
            build is not None
            and build.docker_image_name == image_uri
            and build.status == ClusterEnvironmentBuildStatus.SUCCEEDED
        ):
            return build.id

        # Still create a new build if the cluster env already exists but the build does not match the image_uri.
        result = self._external_api_client.create_cluster_environment_build(
            CreateClusterEnvironmentBuild(
                # For historical reasons, we have to use docker_image_name instead of image_uri; but it is just a URI to the image.
                cluster_environment_id=cluster_env.id,
                docker_image_name=image_uri,
            )
        ).result

        assert result.completed
        return result.cluster_environment_build_id

    def send_workspace_notification(
        self, notification: WorkspaceNotification,
    ):
        if not self.inside_workspace():
            return

        try:
            r = requests.post(WORKSPACE_NOTIFICATION_ADDRESS, json=notification.dict())
            r.raise_for_status()
        except Exception:
            logger.exception(
                "Failed to send workspace notification. "
                "This should not happen, so please contact Anyscale support."
            )

    def get_service(self, name: str) -> Optional[ServiceModel]:
        # TODO(edoakes): this endpoint is very slow and there's no reason we should need
        # to use this complex list endpoint just to fetch a service by name.
        paging_token = None
        project_id = self.get_project_id()
        service: Optional[ServiceModel] = None
        while True:
            resp = self._external_api_client.list_services(
                project_id=project_id,
                name=name,
                count=self.LIST_ENDPOINT_COUNT,
                paging_token=paging_token,
            )
            for result in resp.results:
                if result.name == name:
                    service = result
                    break

            paging_token = resp.metadata.next_paging_token
            if service is not None or paging_token is None:
                break

        return service

    def rollout_service(self, model: ApplyServiceModel) -> ServiceModel:
        result: ServiceModel = self._external_api_client.rollout_service(model).result
        return result

    def rollback_service(
        self, service_id: str, *, max_surge_percent: Optional[int] = None
    ) -> ServiceModel:
        result: ServiceModel = self._external_api_client.rollback_service(
            service_id,
            rollback_service_model=RollbackServiceModel(
                max_surge_percent=max_surge_percent
            ),
        )
        return result

    def terminate_service(self, service_id: str) -> ServiceModel:
        result: ServiceModel = self._external_api_client.terminate_service(service_id)
        return result

    def submit_job(self, model: CreateInternalProductionJob) -> InternalProductionJob:
        job: InternalProductionJob = self._internal_api_client.create_job_api_v2_decorated_ha_jobs_create_post(
            model,
        ).result
        return job

    def upload_local_dir_to_cloud_storage(
        self, local_dir: str, *, cloud_id: str, excludes: Optional[List[str]] = None
    ) -> str:
        if not pathlib.Path(local_dir).is_dir():
            raise RuntimeError(f"Path '{local_dir}' is not a valid directory.")

        with zip_local_dir(local_dir, excludes=excludes) as (
            _,
            zip_file_bytes,
            content_hash,
        ):
            request = CloudDataBucketPresignedUploadRequest(
                file_type=CloudDataBucketFileType.RUNTIME_ENV_PACKAGES,
                file_name=RUNTIME_ENV_PACKAGE_FORMAT.format(content_hash=content_hash),
            )
            info: CloudDataBucketPresignedUploadInfo = self._internal_api_client.generate_cloud_data_bucket_presigned_upload_url_api_v2_clouds_cloud_id_generate_cloud_data_bucket_presigned_upload_url_post(
                cloud_id, request
            ).result
            requests.put(info.upload_url, data=zip_file_bytes).raise_for_status()

        return info.file_uri



from typing import Optional

from anyscale._private.sdk import inject_sdk_singleton
from anyscale.service._private.service_sdk import ServiceSDK
from anyscale.service.models import ServiceConfig, ServiceStatus


_SERVICE_SDK_SINGLETON_KEY = "service_sdk"


@inject_sdk_singleton(_SERVICE_SDK_SINGLETON_KEY, ServiceSDK)
def deploy(
    config: ServiceConfig,
    *,
    in_place: bool = False,
    canary_percent: Optional[int] = None,
    max_surge_percent: Optional[int] = None,
    _sdk: ServiceSDK,
):
    return _sdk.deploy(
        config,
        in_place=in_place,
        canary_percent=canary_percent,
        max_surge_percent=max_surge_percent,
    )


@inject_sdk_singleton(_SERVICE_SDK_SINGLETON_KEY, ServiceSDK)
def rollback(
    name: Optional[str], *, max_surge_percent: Optional[int] = None, _sdk: ServiceSDK,
):
    return _sdk.rollback(name=name, max_surge_percent=max_surge_percent)


@inject_sdk_singleton(_SERVICE_SDK_SINGLETON_KEY, ServiceSDK)
def terminate(name: Optional[str], *, _sdk: ServiceSDK):
    return _sdk.terminate(name=name)


@inject_sdk_singleton(_SERVICE_SDK_SINGLETON_KEY, ServiceSDK)
def status(name: Optional[str], *, _sdk: ServiceSDK) -> ServiceStatus:
    return _sdk.status(name=name)



import copy
from typing import Any, Dict, Optional

from anyscale._private.sdk.workload_sdk import WorkloadSDK
from anyscale.cli_logger import BlockLogger
from anyscale.sdk.anyscale_client.models import (
    AccessConfig,
    ApplyServiceModel,
    ProductionServiceV2VersionModel,
    RayGCSExternalStorageConfig,
    ServiceConfig as ExternalAPIServiceConfig,
    ServiceEventCurrentState,
    ServiceModel,
)
from anyscale.service.models import (
    ServiceConfig,
    ServiceState,
    ServiceStatus,
    ServiceVersionStatus,
)
from anyscale.utils.workspace_notification import (
    WorkspaceNotification,
    WorkspaceNotificationAction,
)


logger = BlockLogger()


class ServiceSDK(WorkloadSDK):
    def _override_application_runtime_envs(
        self,
        config: ServiceConfig,
        *,
        autopopulate_in_workspace: bool = True,
        cloud_id: Optional[str] = None,
    ) -> ServiceConfig:
        """Overrides the runtime_env of each application in the config.

        Local directories specified in the 'working_dir' or 'py_modules' fields will be
        uploaded and replaced with the resulting remote URIs.

        Requirements files will be loaded and populated into the 'pip' field.

        If autopopulate_from_workspace is passed and this code is running inside a
        workspace, the following defaults will be applied:
            - 'working_dir' will be set to '.'.
            - 'pip' will be set to the workspace-managed requirements file.
        """
        new_applications = copy.deepcopy(config.applications)
        new_runtime_envs = [app.get("runtime_env", {}) for app in new_applications]

        new_runtime_envs = self.override_and_upload_local_dirs(
            new_runtime_envs,
            working_dir_override=config.working_dir,
            excludes_override=config.excludes,
            cloud_id=cloud_id,
            autopopulate_in_workspace=autopopulate_in_workspace,
        )
        new_runtime_envs = self.override_and_load_requirements_files(
            new_runtime_envs,
            requirements_override=config.requirements,
            autopopulate_in_workspace=autopopulate_in_workspace,
        )

        for app, new_runtime_env in zip(new_applications, new_runtime_envs):
            if new_runtime_env:
                app["runtime_env"] = new_runtime_env

        return config.options(
            applications=new_applications,
            requirements=None,
            working_dir=None,
            excludes=None,
        )

    def _get_default_name(self) -> str:
        """Get a default name for the service.

        A default is currently only generated when running inside a workspace
        (from the workspace cluster name), so this function errors if called outside
        a workspace.
        """
        name = self.get_current_workspace_name()
        if name is None:
            raise ValueError(
                "A service name must be provided when running outside of a workspace."
            )

        self.logger.info(f"No name was specified, using default: '{name}'.")
        return name

    def _log_deployed_service_info(
        self, service: ServiceModel, *, canary_percent: Optional[int]
    ):
        """Log user-facing information about a deployed service."""
        canary_percent_info = (
            ""
            if canary_percent is None
            else f" (target canary percent: {canary_percent})"
        )
        message = f"Service '{service.name}' deployed{canary_percent_info}."
        self.logger.info(message)
        if self.client.inside_workspace():
            self.client.send_workspace_notification(
                WorkspaceNotification(
                    body=message,
                    action=WorkspaceNotificationAction(
                        type="navigate-service", title="View Service", value=service.id,
                    ),
                ),
            )

        self.logger.info(
            f"View the service in the UI: '{self.client.get_service_ui_url(service.id)}'"
        )
        self.logger.info(
            "Query the service once it's running using the following curl command:"
        )
        auth_token_header = (
            ""
            if service.auth_token is None
            else f"-H 'Authorization: Bearer {service.auth_token}' "
        )
        self.logger.info(f"curl {auth_token_header}{service.base_url}/")

    def _build_apply_service_model_from_config(
        self,
        name: str,
        config: ServiceConfig,
        *,
        in_place: bool = False,
        canary_percent: Optional[int] = None,
        max_surge_percent: Optional[int] = None,
    ) -> ApplyServiceModel:
        if config.containerfile is not None:
            build_id = self.client.get_cluster_env_build_id_from_containerfile(
                cluster_env_name=f"image-for-{name}",
                containerfile=self.get_containerfile_contents(config.containerfile),
            )
        elif config.image_uri is not None:
            build_id = self.client.get_cluster_env_build_id(image_uri=config.image_uri,)
        else:
            build_id = self.client.get_default_build_id()

        compute_config_id = self.resolve_compute_config_to_id(config.compute_config)

        # If a compute config was specified, we need to make sure to used the correct
        # cloud_id when uploading local directories.
        cloud_id = self.client.get_cloud_id(
            compute_config_id=None
            if config.compute_config is None
            else compute_config_id
        )
        config = self._override_application_runtime_envs(config, cloud_id=cloud_id)

        ray_serve_config: Dict[str, Any] = {"applications": config.applications}
        if config.http_options:
            ray_serve_config["http_options"] = config.http_options
        if config.grpc_options:
            ray_serve_config["grpc_options"] = config.grpc_options
        if config.logging_config:
            ray_serve_config["logging_config"] = config.logging_config

        ray_gcs_external_storage_config = None
        if config.ray_gcs_external_storage_config is not None:
            ray_gcs_external_storage_config = RayGCSExternalStorageConfig(
                **config.ray_gcs_external_storage_config
            )

        return ApplyServiceModel(
            name=name,
            project_id=self.client.get_project_id(),
            ray_serve_config=ray_serve_config,
            build_id=build_id,
            compute_config_id=compute_config_id,
            canary_percent=canary_percent,
            max_surge_percent=max_surge_percent,
            rollout_strategy="IN_PLACE" if in_place else "ROLLOUT",
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=config.query_auth_token_enabled),
            ),
            ray_gcs_external_storage_config=ray_gcs_external_storage_config,
        )

    def deploy(
        self,
        config: ServiceConfig,
        *,
        in_place: bool = False,
        canary_percent: Optional[int] = None,
        max_surge_percent: Optional[int] = None,
    ):
        if not isinstance(in_place, bool):
            raise TypeError("in_place must be a bool.")

        if canary_percent is not None:
            if not isinstance(canary_percent, int):
                raise TypeError("canary_percent must be an int.")
            if canary_percent < 0 or canary_percent > 100:
                raise ValueError("canary_percent must be between 0 and 100.")

        if max_surge_percent is not None:
            if not isinstance(max_surge_percent, int):
                raise TypeError("max_surge_percent must be an int.")

            if max_surge_percent < 0 or max_surge_percent > 100:
                raise ValueError("max_surge_percent must be between 0 and 100.")

        name = config.name or self._get_default_name()

        existing_service: Optional[ServiceModel] = self.client.get_service(name)
        if existing_service is None:
            self.logger.info(f"Starting new service '{name}'.")
        elif existing_service.current_state == ServiceEventCurrentState.TERMINATED:
            self.logger.info(f"Restarting existing service '{name}'.")
        else:
            self.logger.info(f"Updating existing service '{name}'.")

        service: ServiceModel = self.client.rollout_service(
            self._build_apply_service_model_from_config(
                name,
                config,
                in_place=in_place,
                canary_percent=canary_percent,
                max_surge_percent=max_surge_percent,
            )
        )

        self._log_deployed_service_info(service, canary_percent=canary_percent)

    def rollback(
        self, name: Optional[str] = None, *, max_surge_percent: Optional[int] = None,
    ):
        if name is None:
            name = self._get_default_name()

        model: Optional[ServiceModel] = self.client.get_service(name)
        if model is None:
            raise RuntimeError(f"Service with name '{name}' was not found.")

        return self.client.rollback_service(
            model.id, max_surge_percent=max_surge_percent,
        )

    def terminate(self, name: Optional[str] = None):
        if name is None:
            name = self._get_default_name()

        model: Optional[ServiceModel] = self.client.get_service(name)
        if model is None:
            raise RuntimeError(f"Service with name '{name}' was not found.")

        return self.client.terminate_service(model.id,)

    def _strip_none_values_from_dict(self, d: Dict) -> Dict:
        """Return a copy of the dictionary without any keys whose values are None.

        Recursively calls into any dictionary values.
        """
        result = {}
        for k, v in d.items():
            if isinstance(v, dict):
                result[k] = self._strip_none_values_from_dict(v)
            elif v is not None:
                result[k] = v

        return result

    def _service_version_model_to_status(
        self,
        model: ProductionServiceV2VersionModel,
        *,
        service_name: str,
        query_auth_token_enabled: bool,
    ) -> ServiceVersionStatus:
        image_uri = self.client.get_cluster_env_build_image_uri(model.build_id)
        if image_uri is None:
            raise RuntimeError(f"Failed to get image URI for ID {model.build_id}.")

        ray_gcs_external_storage_config = None
        if model.ray_gcs_external_storage_config is not None:
            ray_gcs_external_storage_config = self._strip_none_values_from_dict(
                model.ray_gcs_external_storage_config.to_dict()
            )

        return ServiceVersionStatus(
            name=model.version,
            # NOTE(edoakes): there is also a "current_weight" field but it does not match the UI.
            weight=model.weight,
            config=ServiceConfig(
                name=service_name,
                applications=model.ray_serve_config["applications"],
                image_uri=image_uri,
                compute_config=self.get_user_facing_compute_config(
                    model.compute_config_id
                ),
                query_auth_token_enabled=query_auth_token_enabled,
                http_options=model.ray_serve_config.get("http_options", None),
                grpc_options=model.ray_serve_config.get("grpc_options", None),
                logging_config=model.ray_serve_config.get("logging_config", None),
                ray_gcs_external_storage_config=ray_gcs_external_storage_config,
            ),
        )

    def _service_model_to_status(self, model: ServiceModel) -> ServiceStatus:
        # TODO(edoakes): for some reason the primary_version is populated
        # when the service is terminated. This should be fixed in the backend.
        is_terminated = model.current_state == ServiceEventCurrentState.TERMINATED

        # TODO(edoakes): this is currently only exposed at the service level in the API,
        # which means that the per-version `query_auth_token_enabled` field will lie if
        # it's changed.
        query_auth_token_enabled = model.auth_token is not None

        primary_version = None
        if not is_terminated and model.primary_version is not None:
            primary_version = self._service_version_model_to_status(
                model.primary_version,
                service_name=model.name,
                query_auth_token_enabled=query_auth_token_enabled,
            )

        canary_version = None
        if not is_terminated and model.canary_version is not None:
            canary_version = self._service_version_model_to_status(
                model.canary_version,
                service_name=model.name,
                query_auth_token_enabled=query_auth_token_enabled,
            )

        # If we add a new state to the backend, old clients may not recognize it.
        # Rather than erroring out and causing old code to crash, return UNKNOWN.
        try:
            state = ServiceState(model.current_state)
        except ValueError:
            state = ServiceState.UNKNOWN
            self.logger.warning(
                f"Got unrecognized state: '{model.current_state}'. "
                "You likely need to update the 'anyscale' package. "
                "If you still see this message after upgrading, contact Anyscale support."
            )

        return ServiceStatus(
            id=model.id,
            name=model.name,
            state=state,
            query_url=model.base_url,
            query_auth_token=model.auth_token,
            primary_version=primary_version,
            canary_version=canary_version,
        )

    def status(self, name: Optional[str] = None) -> ServiceStatus:
        if name is None:
            name = self._get_default_name()

        model: Optional[ServiceModel] = self.client.get_service(name)
        if model is None:
            raise RuntimeError(f"Service with name '{name}' was not found.")

        return self._service_model_to_status(model)
